#!/usr/bin/python

#/*******************************************************************************
# Copyright 2009 Seapine Software, Inc. All rights reserved.
# 
# IMPORTANT: This Software is supplied to you by Seapine Software, Inc. (Seapine)
# in consideration of your agreement to the following terms, and your use,
# installation, modification or redistribution of this Software constitutes
# acceptance of these terms. If you do not agree with these terms, please do not
# use, install, modify or redistribute this Software.
# 
# In consideration of your agreement to abide by the following terms, and subject
# to these terms, Seapine grants you a personal, non-exclusive license, under
# Seapine copyrights in this original Seapine software (the Software), to use,
# reproduce, modify and redistribute the Software, with or without modifications,
# in source and/or binary forms; provided that if you redistribute the Software
# in its entirety and without modifications, you must retain this notice and the
# following text and disclaimers in all such redistributions of the Software.
# Neither the name, trademarks, service marks or logos of Seapine, Inc. may be
# used to endorse or promote products derived from the Software without specific
# prior written permission from Seapine. Except as expressly stated in this
# notice, no other rights or licenses, express or implied, are granted by Seapine
# herein, including but not limited to any patent rights that may be infringed
# by your derivative works or by other works in which the Software may be
# incorporated.
# 
# The Software is provided by Seapine on an "AS IS" basis. SEAPINE MAKES NO
# WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
# WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE, REGARDING THE SOFTWARE OR ITS USE AND OPERATION ALONE OR IN
# COMBINATION WITH YOUR PRODUCTS. 
# 
# IN NO EVENT SHALL SEAPINE BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) ARISING IN ANY WAY OUT OF THE USE, REPRODUCTION, MODIFICATION
# AND/OR DISTRIBUTION OF THE SOFTWARE, HOWEVER CAUSED AND WHETHER UNDER THEORY
# OF CONTRACT, TORT (INCLUDING NEGLIGENCE), STRICT LIABILITY OR OTHERWISE, EVEN
# IF SEAPINE HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************/

import os
import sys
import PySSCM
import optparse
import tempfile
import pysvn
import time
import operator
import urllib

def rmdir_recursive(dir):
	"""Remove a directory, and all its contents if it is not already empty."""
	try:
		# on Windows, if we don't have write permission we can't remove
		# the file/directory either, so turn that on
		for root, dirs, files in os.walk(dir, topdown=False):
			for name in files:
				full_name=os.path.join(root, name)
				if not os.access(full_name, os.W_OK):
					os.chmod(full_name, 0600)
				os.remove(full_name)
			for name in dirs:
				full_name=os.path.join(root, name)
				if not os.access(full_name, os.W_OK):
					os.chmod(full_name, 0600)
				os.rmdir(full_name)
		os.rmdir(dir)
	except Exception:
		#ignore errors as this is only called from our finally handler
		pass

class SVNChanges(object):
	"""This represents a single SVN change set."""
	
	def __init__(self,svnClient,baseRepo,svnWD,scm,surroundWD,mainlineName,
								scmUn,svnURL,useLabels,useChangelists,svnDict):
		self.svnClient = svnClient
		self.scm=scm
		self.baseRepo = baseRepo
		self.svnWD = svnWD
		self.surroundWD = surroundWD
		self.mainlineName = mainlineName
		self.scmUn = scmUn
		self.svnURL = svnURL
		self.useLabels=useLabels
		self.useChangelists = useChangelists
		self.svnDict = svnDict
		self.newFiles=[]
		self.newRepos=[]
		self.branchs=[]
		self.deletes=[]
		self.modifies=[]
		self.tempWDs=[]

	def parseLog(self,rev_data):
		"""Parse the log entries for this revision"""
		
		if self.useChangelists:
			self.changelistName="SVN:"+str(rev_data.revision)
		else:
			self.changelistName=None
		self.revision = rev_data.revision
		revDate = time.ctime(rev_data.date)
		if rev_data.has_key('author'):
			author=rev_data.author
		else:
			author='no author'
		self.standardComment = "r%d %s %s %s" % (rev_data.revision.number,
												author,revDate,rev_data.message)
		
		print 'Doing revision '+str(rev_data.revision)
		
		#Sort the changed in path order, so we get any parent
		#repos added before any child items
		rev_data.changed_paths.sort(key=operator.itemgetter('path'))
		for rev_change in rev_data.changed_paths:
			rev_change.splitPath= rev_change.path.split('/')
			if rev_change.copyfrom_path is None:
				rev_change.message = self.standardComment
			else:
				rev_change.message = "r%d %s %s (from %s:%d) %s" % (
									rev_data.revision.number,author,revDate,
									str(rev_change.copyfrom_path),
									rev_change.copyfrom_revision.number,
									str(rev_data.message))
			
			if rev_change.action == 'D':
				self.deletes.append(rev_change.path)
			else:
				if rev_change.path not in self.svnDict:
					self.__updatePath(rev_data.revision,rev_change.path)
				currentItem=self.svnDict[rev_change.path]
				if currentItem.good:
					if rev_change.action == 'A':
						#Sometimes we see adds for items which already exist. For files,
						#treat this as a modify. For repos, delete the "old" one and
						#then re-add it
						if currentItem.kind == pysvn.node_kind.dir:
							if currentItem.rev.number != rev_data.revision.number:
								self.__deleteFile(rev_change.path)
								if self.useChangelists:
									self.__commitChangelist()
								self.__updatePath(rev_data.revision,rev_change.path)
							self.newRepos.append(rev_change)
						elif currentItem.kind == pysvn.node_kind.file:
							if currentItem.rev.number == rev_data.revision.number:
								self.newFiles.append(rev_change)
							else:
								self.modifies.append(rev_change)
					elif (rev_change.action == 'M' or rev_change.action == 'R'):
						if currentItem.kind == pysvn.node_kind.file:
							self.modifies.append(rev_change)
	
	def doUpdates(self):
		"""Do the actual actions to the Surround database"""
		
		for repoChange in self.newRepos:
			self.__addRepo(repoChange)

		for newFile in self.newFiles:
			self.__addFile(newFile)

		for modFile in self.modifies:
			self.__modifyFile(modFile)
			
		for oldFile in self.deletes:
			self.__deleteFile(oldFile)
			
		if self.useChangelists:
			self.__commitChangelist()
		for wd in self.tempWDs:
			rmdir_recursive(wd)
		for entry in os.listdir(self.svnWD):
			fullPath=os.path.join(self.svnWD,entry)
			if os.path.isfile(fullPath):
				os.remove(fullPath)
		if self.useLabels:
			labelName="SVN:"+str(self.revision)
			self.scm.label("/",labelName,Branch=self.mainlineName,
										Repository=self.baseRepo,Recursive=True)


	def __updatePath(self,revision,path):
		"""Update our dictionary of SVN items with any new ones"""
		
		#returns a list of tuples (item name,psvninfo). URL encode the path
		goodPath=self.svnURL+'/'.join([urllib.quote(item) for item in path.split('/')])
		svnList=self.svnClient.info2(goodPath,revision=self.revision,
										peg_revision=self.revision)

		#the first item is the path requesed
		svnList[0][1].good=True
		self.svnDict[path]=svnList[0][1]
		
		#Now get any child items, if there are any
		for svnItem in svnList[1:]:
			fullPath=path+'/'+svnItem[0]
			svnItem[1].good=True
			self.svnDict[fullPath]=svnItem[1]
						
	def __getParentRepo(self,splitPath):
		return self.baseRepo+'/'.join(splitPath[:-1])

	def __modifyFile(self,fileChange):
		newRepo = self.__getParentRepo(fileChange.splitPath)
		surroundPath=os.path.join(self.surroundWD,*fileChange.splitPath)
			
		fileName=fileChange.splitPath[-1]
		self.scm.checkout(fileName,Branch=self.mainlineName,Repository=newRepo)
		self.svnClient.export(self.svnURL+fileChange.path,surroundPath,
					revision=self.revision,peg_revision=self.revision,force=True)
		self.scm.checkin(fileName,Branch=self.mainlineName,
							Comment=fileChange.message,Repository=newRepo,
							Update=True,Changelist=self.changelistName)
	
	def __deleteFile(self,oldFile):
		try:
			path= oldFile.split('/')
			parentRepo = self.__getParentRepo(path)
			self.scm.rm(path[-1],Branch=self.mainlineName,Repository=parentRepo,
						Comment=self.standardComment,Force=True,
						Changelist=self.changelistName,Destroy=True)
			surroundPath=os.path.join(self.surroundWD,*path)
			if os.path.isfile(surroundPath):
				os.remove(surroundPath)
			elif os.path.isdir(surroundPath):
				rmdir_recursive(surroundPath)
		except Exception:
			#Ignore errors when deleting, as the file may not have been created
			pass
		if oldFile in self.svnDict:
			if self.svnDict[oldFile].kind == pysvn.node_kind.dir:
				self.__deleteRepoFromDict(oldFile)
			else:
				del self.svnDict[oldFile]
	
	def __deleteRepoFromDict(self,oldItem):
		if oldItem in self.svnDict and self.svnDict[oldItem].kind == pysvn.node_kind.dir:
			rootPath=oldItem+'/'
			childItems=[anItem for anItem in self.svnDict.keys() if anItem.startswith(rootPath)]
			for childItem in childItems:
				del self.svnDict[childItem]
			del self.svnDict[oldItem]

	def __addFile(self,newFile):
		newRepo = self.__getParentRepo(newFile.splitPath)

		if newFile.copyfrom_path is not None:
			sourcePath=newFile.copyfrom_path.split('/')
			#check to see if the copy from revision is one back,
			#so we can do a share/break share to simulate a copy
			if (newFile.copyfrom_revision.number+1) == self.revision.number:
				# a rename is a add and a remove, check for that case
				if newFile.splitPath[:-1] == sourcePath[:-1] and \
					newFile.copyfrom_path in self.deletes:
					self.scm.rename(sourcePath[-1],newFile.splitPath[-1],
									Repository=newRepo,Branch=self.mainlineName,
									Comment=newFile.message,
									Changelist=self.changelistName)
					self.deletes.remove(newFile.copyfrom_path)
					if newFile.copyfrom_path in self.svnDict:
						del self.svnDict[newFile.copyfrom_path]
				else:
					# This will be the name of the shared file, before we rename it
					tempName='/'.join(newFile.splitPath[:-1]+sourcePath[-1:])
					if tempName not in self.svnDict:
						if self.useChangelists:
							#check to see if the file is coming into a newly
							#added repo. If so, we need to commit the changelist
							#so Surround knows about it before we do a share/break share
							for testRepo in self.newRepos:
								if newFile.path.startswith(testRepo.path):
									self.__commitChangelist()
									self.newRepos=[]
									break
						sourceRepo=self.baseRepo+'/'.join(sourcePath[:-1])
						self.scm.share(sourceRepo,[sourcePath[-1]],
										Branch=self.mainlineName,
										Comment=newFile.message,
										DestinationRepository=newRepo)
						self.scm.breakshare([sourcePath[-1]],
											Branch=self.mainlineName,
											Comment=newFile.message,
											Repository=sourceRepo,All=True)
						if newFile.splitPath[-1] != sourcePath[-1]:
							self.scm.rename(sourcePath[-1],newFile.splitPath[-1],
											Repository=newRepo,
											Branch=self.mainlineName,
											Comment=newFile.message,
											Changelist=self.changelistName)
					else:
						# The share name would exist before we can rename it,
						# so we just have to do an add 
						self.__addHelper(newFile,newRepo)
			else:
				#Copying from an old version
				self.__addHelper(newFile,newRepo)
		else:
			#Adding a brand new file
			self.__addHelper(newFile,newRepo)

	def __addHelper(self,newFile,newRepo):
		fullPath=os.path.join(self.svnWD,newFile.splitPath[-1])
		try:
			self.svnClient.export(self.svnURL+newFile.path,fullPath,
									revision=self.revision,
									peg_revision=self.revision,force=True)
		except Exception:
			#Files may have illegal names
			self.svnDict[newFile.path].good=False
		else:
			self.scm.add(fullPath,Branch=self.mainlineName,
						Comment=newFile.message,Repository=newRepo,Restore=True,
						Changelist=self.changelistName)

	def __addRepo(self,repoChange):
		parentRepo = self.__getParentRepo(repoChange.splitPath)
		newRepo=repoChange.splitPath[-1]

		if repoChange.copyfrom_path is None:
			self.scm.mkrepository(newRepo,Branch=self.mainlineName,
								Comment=repoChange.message,Repository=parentRepo)
		else:
			#check to see if the copy from revision is one back,
			#so we can do a share/break share to simulate a copy
			if (repoChange.copyfrom_revision.number+1) == self.revision.number:
				sourcePath=repoChange.copyfrom_path.split('/')
				# a rename is a add and a remove, check for that case
				if repoChange.splitPath[:-1] == sourcePath[:-1] and \
					repoChange.copyfrom_path in self.deletes:
					self.scm.rename(sourcePath[-1],newRepo,Repository=parentRepo,
									Branch=self.mainlineName,
									Comment=repoChange.message,
									Changelist=self.changelistName)
					self.deletes.remove(repoChange.copyfrom_path)
					if repoChange.copyfrom_path in self.svnDict:
						self.__deleteRepoFromDict(repoChange.copyfrom_path)
				else:
					self.__shareRepo(repoChange,sourcePath,parentRepo,newRepo)
			else:
				#Sharing from some prior revision, so just treat this as an add
				copyWD = tempfile.mkdtemp()
				self.tempWDs.append(copyWD)
				fullPath=os.path.join(copyWD,newRepo)
				self.svnClient.export(self.svnURL+repoChange.copyfrom_path,
										fullPath,
										revision=repoChange.copyfrom_revision,
										peg_revision=repoChange.copyfrom_revision,
										ignore_externals=True)
				self.scm.add([fullPath],Branch=self.mainlineName,
								Comment=repoChange.message,Repository=parentRepo,
								Restore=True,Recursive=True,Get=True,
								Changelist=self.changelistName)
	
	def __shareRepo(self,repoChange,sourcePath,parentRepo,newRepo):
		sourceRepo=self.baseRepo+'/'.join(sourcePath)
		#Check to see if the source of the copy is in the same parent repo.
		# If so, can't just share the source to the destination we need to
		# share all the sub-elements
		if repoChange.splitPath[:-1] == sourcePath[:-1]:
			self.scm.mkrepository(newRepo,Branch=self.mainlineName,
									Comment=repoChange.message,
									Repository=parentRepo)
			repoList=self.scm.ls('/',Branch=self.mainlineName,
								Repository=sourceRepo,Recursive=True)
			
			newParent = parentRepo+"/"+newRepo
			if sourceRepo in repoList and len(repoList[sourceRepo]) > 0:
				self.scm.share(sourceRepo,Files=repoList[sourceRepo],
								Branch=self.mainlineName,
								Comment=repoChange.message,
								DestinationRepository=newParent)
			self.scm.get(['/'],Branch=self.mainlineName,Repository=sourceRepo)
			sourceDirectory = os.path.join(self.surroundWD,*sourcePath)
			for entry in os.listdir(sourceDirectory):
				if os.path.isdir(os.path.join(sourceDirectory,entry)):
					self.scm.share(sourceRepo+'/'+entry,
									Branch=self.mainlineName,
									Comment=repoChange.message,
									DestinationRepository=newParent,
									Recursive=True)
			for repoKey, fileItem in repoList.items():
				if len(fileItem) > 0:
					self.scm.breakshare(fileItem,
										Branch=self.mainlineName,
										Comment=repoChange.message,
										Repository=repoKey,All=True)
		else:
			shareName='/'.join(repoChange.splitPath[:-1]+sourcePath[-1:])
			self.scm.share(sourceRepo,Branch=self.mainlineName,
							Comment=repoChange.message,
							DestinationRepository=parentRepo,
							Recursive=True)
			repoList=self.scm.ls('/',Branch=self.mainlineName,
								Repository=sourceRepo,Recursive=True)
			for repoKey, fileItem in repoList.items():
				if len(fileItem) > 0:
					self.scm.breakshare(fileItem,
										Branch=self.mainlineName,
										Comment=repoChange.message,
										Repository=repoKey,All=True)
	
			if repoChange.splitPath[-1] != sourcePath[-1]:
				self.scm.rename(sourcePath[-1],newRepo,
								Repository=parentRepo,
								Branch=self.mainlineName,
								Comment=repoChange.message,
								Changelist=self.changelistName)

	def __commitChangelist(self):
		if self.useChangelists:
			changelists=self.scm.lschangelist(Mainline=self.mainlineName,
												User=self.scmUn)
			if self.changelistName in changelists:
				scm.commitchangelist(changelists[self.changelistName][0],
										Mainline=self.mainlineName)


def main():
	options = GetOptions()
	
	scm = PySSCM.SSCM()
	scm.Username = options.scmUn
	scm.Password=options.scmPS
	scm.SCMServerAddr=options.scmAddress
	scm.PortNum=options.scmPort
	
	client = pysvn.Client()
	client.exception_style=1
	client.callback_ssl_server_trust_prompt = lambda trust_dict: (True, 0, True)
	info=client.info2(options.svnURL,recurse=False)[0][1]
	#info return a list of tuples, the info we want is the
	#second entry in the first tuple
	
	#Get the new Surround mainline name
	if options.mainlineName is None:
		options.mainlineName = info[0][1].repos_root_URL.rsplit('/',1)[1]
			
		if len(options.mainlineName) == 0:
			options.mainlineName='SVN Import'
	
	if options.surroundRepository is None:
		options.surroundRepository = options.mainlineName
	
	mainlines = [mainline.strip() for mainline in scm.lsmainline()]
	if options.mainlineName not in mainlines:
		scm.mkmainline(options.mainlineName)
		
	svnWD = tempfile.mkdtemp()
	surroundWD = None
	try:
		baseRepo=options.surroundRepository
		surroundWD=tempfile.mkdtemp()
		scm.workdir(surroundWD,Repository=baseRepo,Branch=options.mainlineName,
					Inherit=True)

		print 'Getting log data'
		log_data=client.log( info.URL,discover_changed_paths=True)
		log_data.reverse()

		svnDict={}
		for rev_data in log_data:
			newChange = SVNChanges(client,baseRepo,svnWD,scm,surroundWD,
									options.mainlineName,options.scmUn,
									options.svnURL,options.useLabels,
									options.useChangelists,svnDict)
			newChange.parseLog(rev_data)
			newChange.doUpdates()
	finally:
		try:
			rmdir_recursive(svnWD)
			if surroundWD is not None:
				rmdir_recursive(surroundWD)
			scm.rmworkdir(Repository=options.surroundRepository,
														Branch=options.branch)
		except Exception:
			pass

def GetOptions():
	parser = optparse.OptionParser()
	parser.add_option("-a", dest="svnUN",
	                  help="Enter the username used to log in to the subversion database",
	                  metavar="SVNUsername")
	parser.add_option("-b", dest="svnPS",
	                  help="Enter the password used to log in to the subversion database",
	                  metavar="SVNPassword")
	parser.add_option("-c", dest="svnURL",
	                  help="Enter the SVN repository URL",
	                  metavar="SVN_URL")
	parser.add_option("-d", dest="svnWD",
	                  help="Enter the SVN local working directory",
	                  metavar="SVN_WD")
	parser.add_option("-e", dest="surroundRepository",
	                  help="Specify the Surround SCM repository to import SourceSafe into",
	                  metavar="SurroundRepository")
	parser.add_option("-f", dest="useLabels",action="store_true", default=False,
	                  help="Create label for each changeset")
	parser.add_option("-i", dest="useChangelists",action="store_true", default=False,
	                  help="Create changelists for each changeset")
	parser.add_option("-g", dest="mainlineName",
	                  help="Enter the name of the Surround SCM mainline branch",
	                  metavar="NewMainlineName")
	parser.add_option("-l", dest="logFile",
	                  help="Specify the file name to log results to",
	                  metavar="logFile")
	parser.add_option("-w", dest="scmUn",
	                  help="Enter the username used to log in to Surround SCM",
	                  metavar="SSCMUsername")
	parser.add_option("-x", dest="scmPS",
	                  help="Enter the password used to log in to Surround SCM",
	                  metavar="SSCMPassword")
	parser.add_option("-y", dest="scmAddress",
	                  help="Enter the address of Surround SCM Server host computer",
	                  metavar="SSCMServerAddr")
	parser.add_option("-z", dest="scmPort",
	                  help="Enter the port number of Surround SCM Server host computer",
	                  metavar="SSCMPort")
	
	(options, args) =parser.parse_args()
	
	options.svnURL= options.svnURL.rstrip('/')
#	options.svnURL = "http://svn.collab.net/repos/svn"
#	options.scmUn = 'administrator'
#	options.scmAddress='localhost'
#	options.scmPort=4900
#	options.mainlineName='Subversion1'
	return options


if __name__ == "__main__":
	 sys.exit(main())